from dataclasses import asdict

import pytest

from idp_authentication import UserRole
from idp_authentication.conftest import TEST_PERMISSION, TEST_ROLE_1, TEST_ROLE_2


@pytest.mark.parametrize(
    "user,user_role,expected_role",
    [
        ("user", "test_role_3", "test_role_3"),
        ("user_2", "test_role_4", "test_role_4"),
    ],
    indirect=["user"],
)
def test_user_role_create(container, user, user_role, expected_role):
    process_user, user_as_dict = process_user_use_case(container, user, user_role)
    new_role = UserRole(role=user_role, app_entities_restrictions={}, user_id=user.id)
    user_as_dict["user_roles"] = new_role
    process_user.execute(user_as_dict)
    assert user.user_roles[0].role == expected_role


@pytest.mark.parametrize(
    "user,user_role,expected_role",
    [
        ("user", TEST_ROLE_1, TEST_ROLE_1),
        ("user_2", TEST_ROLE_2, TEST_ROLE_2),
    ],
    indirect=["user"],
)
def test_user_role_update(container, user, user_role, expected_role):
    process_user, user_as_dict = process_user_use_case(container, user, user_role)
    process_user.execute(user_as_dict)
    assert user.user_roles[0].role == expected_role


def process_user_use_case(container, user, user_role):
    process_user = container.users_module().use_cases().process_user_message_use_case()
    user_as_dict = asdict(user)
    user_as_dict["app_specific_configs"] = {
        "test": {
            "default": {
                user_role: {
                    "app_entities_restrictions": {
                        "vehicle": [3],
                    },
                    "permission_restrictions": {TEST_PERMISSION: {"vehicle": [1, 2]}},
                }
            }
        }
    }
    return process_user, user_as_dict



"""IDP authentication package. Hexagonal architecture implementation."""

__version__ = "0.0.9"

from .users.adapters import *
from .users.base_classes import *
from .users.domain.entities import *
from .users.domain.ports import *
from .users.domain.use_cases import *



import pytest

from idp_authentication.users.tests.mock_event_producer import MockEventProducer


@pytest.mark.parametrize(
    "deleted",
    [
        True,
        False,
    ],
)
def test_create_or_delete_app_entity_record_event(container, vehicle, deleted):
    # Arrange
    event_producer = MockEventProducer()

    # Act
    with container.event_producer.override(event_producer):
        use_case = (
            container.users_module.use_cases.send_app_entity_record_event_use_case.provided()
        )
        use_case.execute(app_entity_record=vehicle, deleted=deleted)

    # Assert
    assert event_producer.value == {
        "app_identifier": container.config.provided().get("APP_IDENTIFIER"),
        "app_entity_type": vehicle.entity_type,
        "record_identifier": vehicle.idp_identifier,
        "label": vehicle.idp_label,
        "deleted": deleted,
    }



from datetime import datetime

from idp_authentication.custom_types import AppEntityEventMessage
from idp_authentication.users.base_classes.base_use_case import UseCasePort
from idp_authentication.users.domain.entities.app_entity import AppEntity
from idp_authentication.users.domain.ports.event_producer import EventProducerPort


class SendAppEntityRecordEventUseCase(UseCasePort):
    def __init__(
        self,
        event_producer: EventProducerPort,
        app_identifier: str,
    ):
        self.event_producer = event_producer
        self.app_identifier = app_identifier

    def execute(self, app_entity_record: AppEntity, deleted: bool = False):
        app_entity_record_event: AppEntityEventMessage = {
            "app_identifier": self.app_identifier,
            "app_entity_type": app_entity_record.entity_type,
            "record_identifier": app_entity_record.idp_identifier,
            "label": app_entity_record.idp_label,
            "deleted": deleted,
        }
        self.event_producer.produce(
            topic="app_entity_record_events",
            key=str(datetime.now()),
            value=app_entity_record_event,
        )



from sqlalchemy import Column, DateTime, func
from sqlalchemy.orm import declarative_base, declarative_mixin

from idp_authentication.users.infrastructure.database.sqlalchemy.custom_orm_column_types import (
    UUID,
)

Base = declarative_base()


@declarative_mixin
class BaseModelMixin:
    id = Column(
        UUID(), primary_key=True, unique=True
    )  # UUID will be used, which has 36 chars.
    created_at = Column(DateTime, default=func.now())
    updated_at = Column(DateTime, default=func.now())



from .authorize_app_entity_records import AuthorizeEntityRecordsUseCase
from .authorize_or_get_all_allowed_record_identifiers import (
    AuthorizeOrGetAllAllowedRecordIdentifiersUseCase,
)
from .create_or_update_user import CreateOrUpdateUserUseCase
from .get_allowed_records_identifiers import GetAllowedRecordIdentifiersUseCase
from .get_user import GetUserUseCase
from .get_users_with_access_to_app_entity_record import (
    GetUsersWithAccessToAppEntityRecordUseCase,
)
from .process_user import ProcessUserMessageUseCase
from .send_app_entity_record_event import SendAppEntityRecordEventUseCase



from typing import ClassVar, Generic, List, Optional

from idp_authentication.custom_types import T
from idp_authentication.users.domain.ports import RepositoryPort, SessionPort


class BaseRepository(Generic[T], RepositoryPort):
    entity: ClassVar[T]

    def __init__(self, session: SessionPort):
        self.session = session

    def get_or_none(self, **filters) -> Optional[T]:
        return self.session.query(self.entity).filter_by(**filters).one_or_none()

    def get_first(self, **filters) -> Optional[T]:
        return self.session.query(self.entity).filter_by(**filters).first()

    def all(self) -> List[T]:
        return self.session.query(self.entity).all()

    def create(self, **attr) -> T:
        record = self.entity(**attr)
        self.session.add(record)
        return record

    def update_record(self, record: T, **attr) -> T:
        for key, value in attr.items():
            setattr(record, key, value)
        return record

    def delete(self, record: T) -> None:
        self.session.delete(record)



import json

from sqlalchemy import String, TypeDecorator
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.dialects.postgresql import UUID as POSTGRES_UUID


class UUID(TypeDecorator):
    impl = String(36)
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(POSTGRES_UUID)
        else:
            return dialect.type_descriptor(self.impl)


class JsonField(TypeDecorator):
    impl = String()
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(JSONB)
        else:
            return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value, dialect):
        if dialect.name == "postgresql":
            return value
        return json.dumps(value) if value else None

    def process_result_value(self, value, dialect):
        if dialect.name == "postgresql":
            return value
        return json.loads(value) if value else None



from collections import defaultdict
from copy import deepcopy
from typing import List

from idp_authentication.custom_types import UserRecordDict
from idp_authentication.enums import ChoiceEnum
from idp_authentication.users.base_classes.base_use_case import UseCasePort
from idp_authentication.users.domain.ports.unit_of_work import UsersUnitOfWorkPort

"""
{
   "username":"test",
   "first_name":"test",
   "last_name":"test",
   "email":"test@cardoai.com",
   "is_active":true,
   "is_staff":true,
   "is_superuser":true,
   "date_joined":"2022-01-01",
   "app_specific_configs":{
      "test":{
         "tenant":{
            "Servicer":{
               "app_entities_restrictions":{
                  "vehicle":[
                     1,
                     2
                  ]
               },
               "permission_restrictions":{
                  "synchronizeDoD":false
               }
            }
         }
      }
   }
}
"""


class ProcessUserMessageUseCase(UseCasePort):
    def __init__(
        self,
        users_unit_of_work: UsersUnitOfWorkPort,
        roles: ChoiceEnum,
        create_or_update_user_use_case: UseCasePort,
        tenants: List[str],
        app_identifier: str,
    ):
        self.users_unit_of_work = users_unit_of_work
        self.roles = roles

        self.data = None
        self.app_identifier = app_identifier
        self.create_or_update_user_user_case = create_or_update_user_use_case
        self.tenants = tenants or ["default"]

    def execute(self, data: UserRecordDict):
        """
        Extract tenants from the user record and call _update_user for each tenant.
        Remove tenant information from the payload of _update_user since it is not needed
        inside it.

        Send signals before and after calling the _update_user method for each tenant
        separately. This gives possibility to the project to react on the user update.

        One case of handling this signal is to switch database connection to the tenant's
        database. In this way the user can be updated in the correct database.

        """
        app_specific_configs = data.get("app_specific_configs", {})
        if self.app_identifier in app_specific_configs:
            self._process_user_data_with_access_to_app(data)
        else:
            # TODO: To Implement
            pass

    def _process_user_data_with_access_to_app(self, data: UserRecordDict):
        reported_user_app_configs = data.get("app_specific_configs", {}).get(
            self.app_identifier, {}
        )
        tenants = reported_user_app_configs.keys()
        for tenant in tenants:
            if tenant not in self.tenants:
                # logger.info(f"Tenant {tenant} not present, skipping.")
                continue

            user_record_for_tenant = deepcopy(data)
            user_record_for_tenant["app_specific_configs"] = reported_user_app_configs[
                tenant
            ]

            self._update_user(user_record_for_tenant)

    def _update_user(self, user_record: UserRecordDict):
        """
        This method makes sure that the changes that are coming from the IDP
        for a user are propagated in the internal product Authorization Schemas

        Step 1: Create or update User Object
        Step 2: Create/Update/Delete User Roles for this user.
        """
        user = self.create_or_update_user_user_case.execute(user_record)

        with self.users_unit_of_work as uow:
            current_user_roles = defaultdict()
            for user_role in user.user_roles:
                current_user_roles[user_role.role] = user_role

            roles_data = user_record.get("app_specific_configs")
            for role, role_data in roles_data.items():
                if existing_user_role := current_user_roles.get(role):
                    uow.user_role_repository.update_record(
                        existing_user_role,
                        permission_restrictions=role_data.get(
                            "permission_restrictions"
                        ),
                        app_entities_restrictions=role_data.get(
                            "app_entities_restrictions"
                        ),
                    )
                else:
                    created_user_role = uow.user_role_repository.create(
                        user_id=user.id,
                        role=role,
                        permission_restrictions=role_data.get(
                            "permission_restrictions"
                        ),
                        app_entities_restrictions=role_data.get(
                            "app_entities_restrictions"
                        ),
                    )
                    user.user_roles.append(created_user_role)

            # Verify if any of the previous user roles is not being reported anymore
            # Delete it if this is the case
            for role, user_role in current_user_roles.items():
                if roles_data.get(role) is None:
                    # Delete Role Object
                    uow.user_role_repository.delete(user_role)
                    # Delete Role from User Object
                    uow.user_repository.remove_user_role(user, user_role)

            uow.commit()
